from .main import (
    JsonRpcPeer,
    JsonRpcRequest,
    JsonRpcResponse,
)
from .exc import (
    JsonRpcApplicationError,
    JsonRpcError,
    JsonRpcException,
    JsonRpcInvalidParamsError,
    JsonRpcInvalidRequestError,
    JsonRpcInternalError,
    JsonRpcMethodNotFoundError,
    JsonRpcReservedError,
    JsonRpcParseError,
)



from __future__ import annotations
from dataclasses import dataclass
import typing


from .types import JsonDict, JsonList, JsonPrimitive


@dataclass
class JsonRpcError:
    """ Represents an error in the JSON RPC protocol. """

    code: int
    message: str
    data: typing.Optional[JsonDict] = None

    def to_json_dict(self) -> JsonDict:
        """ Convert to a JSON dictionary. """
        dict_ = typing.cast(JsonDict, {"code": self.code, "message": self.message})
        if self.data is not None:
            dict_["data"] = self.data
        return dict_

    @classmethod
    def from_json_dict(cls, json_dict: JsonDict) -> JsonRpcError:
        """ Return a new response from a JSON dictionary. """
        return cls(
            code=typing.cast(int, json_dict["code"]),
            message=typing.cast(str, json_dict["message"]),
            data=typing.cast(typing.Optional[JsonDict], json_dict.get("data")),
        )

    def __repr__(self) -> str:
        """ Return string representation. """
        data = repr(self.data)
        message = repr(self.message)
        return f"JsonRpcError(code={self.code}, message={message}, data={data})"


class JsonRpcException(Exception):
    """
    A base class for JSON-RPC exceptions.

    This exception is never thrown by ``sansio-jsonrpc`` but since it is an ancestor for
    JSON-RPC exceptions, it is useful in `try/except` blocks as a catch-all. You
    generally should not instantiate this class yourself. Instead, use
    class:`JsonRpcApplicationException`, either directly or by creating your own
    subclass of it.
    """

    def __init__(self, error: JsonRpcError):
        """ Constructor. """
        self._error: JsonRpcError = error

    @property
    def code(self):
        """ The JSON-RPC error code. """
        return self._error.code

    @property
    def message(self):
        """ A JSON-RPC error message. """
        return self._error.message

    @property
    def data(self) -> typing.Optional[JsonDict]:
        """ Arbitrary data attached to the error. """
        return self._error.data

    def get_error(self) -> JsonRpcError:
        """ Return the error underlying this exception. """
        return self._error

    def __repr__(self):
        cls = self.__class__.__name__
        code = self._error.code
        data = repr(self._error.data)
        message = repr(self._error.message)
        return f"{cls}<code={code}, message={message}, data={data}>"

    @staticmethod
    def exc_from_error(error: JsonRpcError) -> JsonRpcException:
        """
        Create a new exception derived from the given error.

        When you receive an error response, you may want to raise a corresponding
        exception. This method finds an appropriate exception subclass that matches the
        error's numeric code and instantiates it. It uses some metaclass black magic so
        that it can even return custom subclasses you define in your own code!
        """
        exc: JsonRpcException
        if -32768 <= error.code <= -32000:
            exc = JsonRpcReservedError.exc_from_error(error)
        else:
            exc = JsonRpcApplicationError.exc_from_error(error)
        return exc


class JsonRpcReservedErrorMeta(type):
    """
    This metaclass builds a map of error codes to classes.

    This is used to find the right class to instantiate when an error response is
    received. It also enforces the requirement that reserved error codes must be in the
    range [-32768, -32000].
    """

    _error_classes: typing.Dict[int, type] = dict()

    def __init__(cls, name, bases, attrs):
        if cls.ERROR_CODE is not None and (
            cls.ERROR_CODE < -32768 or cls.ERROR_CODE > -32000
        ):
            raise RuntimeError(
                "Subclasses of JsonRpcReservedError must set ERROR_CODE in range "
                "[-32768, -32000]."
            )
        JsonRpcReservedErrorMeta._error_classes[cls.ERROR_CODE] = cls


class JsonRpcReservedError(JsonRpcException, metaclass=JsonRpcReservedErrorMeta):
    """
    An exception corresponding to the range of codes reserved by the spec.

    The error code must be in the range [-32768, -32000].
    """

    ERROR_CODE: int = -32000
    ERROR_MESSAGE: str = "JSON-RPC reserved error"

    def __init__(
        self,
        message: typing.Optional[str] = None,
        data: typing.Optional[JsonDict] = None,
    ):
        error = JsonRpcError(self.ERROR_CODE, message or self.ERROR_MESSAGE, data)
        super().__init__(error)

    @staticmethod
    def exc_from_error(error: JsonRpcError) -> JsonRpcReservedError:
        """
        Create a new reserved exception that corresponds to the error code.

        This first searches for a subclass that is registered with the given error code.
        If it does not find such a subclass, then it returns a ``JsonRpcReservedError``
        instead.
        """
        cls = JsonRpcReservedError._error_classes.get(error.code, JsonRpcReservedError)
        return cls(error.message, error.data)


class JsonRpcApplicationErrorMeta(type):
    """
    This metaclass builds a map of error codes to classes.

    This is used to find the right class to instantiate when an error response is
    received. It also enforces the requirement that application error codes must **not**
    be in the range [-32768, -32000].
    """

    _error_classes: typing.Dict[int, type] = dict()

    def __init__(cls, name, bases, attrs):
        if (
            cls.ERROR_CODE is not None
            and cls.ERROR_CODE >= -32768
            and cls.ERROR_CODE <= -32000
        ):
            raise RuntimeError(
                "Subclasses of JsonRpcReservedError must set ERROR_CODE outside the "
                " range [-32768, -32000]."
            )
        JsonRpcApplicationErrorMeta._error_classes[cls.ERROR_CODE] = cls


class JsonRpcApplicationError(JsonRpcException, metaclass=JsonRpcApplicationErrorMeta):
    """
    An exception corresponding to the unreserved range of error codes.

    The error code must **not** be in the range [-32768, -32000].
    """

    ERROR_CODE: int = -1
    ERROR_MESSAGE: str = "JSON-RPC"

    def __init__(
        self,
        message: typing.Optional[str] = None,
        *,
        data: typing.Optional[JsonDict] = None,
        code: typing.Optional[int] = None,
    ):
        error = JsonRpcError(
            code or self.ERROR_CODE, message or self.ERROR_MESSAGE, data
        )
        super().__init__(error)

    @staticmethod
    def exc_from_error(error: JsonRpcError) -> JsonRpcApplicationError:
        cls = JsonRpcApplicationError._error_classes.get(
            error.code, JsonRpcApplicationError
        )
        return cls(error.message, data=error.data)


class JsonRpcParseError(JsonRpcReservedError):
    """ Invalid JSON was received by the server. """

    ERROR_CODE = -32700
    ERROR_MESSAGE = "Invalid JSON was received by the server."


class JsonRpcInvalidRequestError(JsonRpcReservedError):
    """ The JSON sent is not a valid Request object. """

    ERROR_CODE = -32600
    ERROR_MESSAGE = "The JSON sent is not a valid Request object."


class JsonRpcMethodNotFoundError(JsonRpcReservedError):
    """
    The method does not exist / is not available.

    This exception is never thrown by ``sansio-jsonrpc``. It should be thrown in
    downstream libraries.
    """

    ERROR_CODE = -32601
    ERROR_MESSAGE = "The method does not exist / is not available."


class JsonRpcInvalidParamsError(JsonRpcReservedError):
    """
    Invalid method parameter(s).

    This exception is never thrown by ``sansio-jsonrpc``. It should be thrown in
    downstream libraries.
    """

    ERROR_CODE = -32602
    ERROR_MESSAGE = "Invalid method parameter(s)."


class JsonRpcInternalError(JsonRpcReservedError):
    """
    Internal JSON-RPC error.
    """

    ERROR_CODE = -32602
    ERROR_MESSAGE = "Internal JSON-RPC error."



from __future__ import annotations
from dataclasses import dataclass
import itertools
import json
import typing

from .exc import (
    JsonRpcError,
    JsonRpcInternalError,
    JsonRpcInvalidRequestError,
    JsonRpcParseError,
)
from .types import (
    JsonDict,
    JsonList,
    JsonPrimitive,
    JsonRpcId,
    JsonRpcParams,
)


class MissingId:
    """ A sentinel class used to indicate that a request is missing an ID. """

    pass


def validate_json_rpc_id(id_: JsonRpcId, exc: typing.Callable):
    """
    Validation routine for the ID field.

    The type of exception raised varies depending on whether we are validating a request
    or a response. This sounds a bit silly, but it is part of the spec.
    """
    id_type = type(id_)
    if id_type is float and not typing.cast(float, id_).is_integer():
        raise exc("`id` number cannot have a fractional part.")
    # Note that isinstance() doesn't work for ID, because int is valid and bool is
    # not, but in Python bool is a subclass of int.
    elif id_type not in (int, float, str):
        raise exc("`id` must be a number, string, or null.")


@dataclass
class JsonRpcRequest:
    """ Represents a JSON RPC request. """

    id: typing.Union[JsonRpcId, MissingId]
    method: str
    params: typing.Optional[JsonRpcParams] = None
    jsonrpc: str = "2.0"

    @property
    def is_notification(self):
        """ True if this request is a notification. """
        return isinstance(self.id, MissingId)

    def __post_init__(self):
        """ Validation logic. """
        if not self.is_notification:
            validate_json_rpc_id(self.id, JsonRpcInvalidRequestError)

        if not isinstance(self.method, str):
            raise JsonRpcInvalidRequestError("`method` must be a string.")

        if not (self.params is None or isinstance(self.params, (dict, list))):
            raise JsonRpcInvalidRequestError("`params` must a list or object.")

        if self.jsonrpc != "2.0":
            raise JsonRpcInvalidRequestError('`jsonrpc` must be "2.0".')

    def to_json_dict(self) -> JsonDict:
        """ Convert to a JSON dictionary. """
        dict_ = typing.cast(JsonDict, {"method": self.method, "jsonrpc": self.jsonrpc})
        if not isinstance(self.id, MissingId):
            dict_["id"] = self.id
        if self.params is not None:
            dict_["params"] = self.params
        return dict_

    @classmethod
    def from_json_dict(cls, json_dict: JsonDict) -> JsonRpcRequest:
        """ Create a new request from a JSON dictionary. """
        id_ = typing.cast(typing.Optional[JsonRpcId], json_dict.get("id", MissingId()))
        params: typing.Optional[JsonDict]
        if "params" in json_dict:
            params = typing.cast(JsonDict, json_dict["params"])
        else:
            params = None
        return cls(
            id=id_,
            method=typing.cast(str, json_dict["method"]),
            params=params,
            jsonrpc=typing.cast(str, json_dict["jsonrpc"]),
        )


# This type is not defined in the types module because it relies on the class above.
JsonRpcRequestHandler = typing.Callable[[JsonRpcRequest], None]


@dataclass
class JsonRpcResponse:
    """ Represents a JSON RPC response. """

    id: JsonRpcId
    result: typing.Optional[JsonPrimitive] = None
    error: typing.Optional[JsonRpcError] = None
    jsonrpc: str = "2.0"

    def __post_init__(self):
        """ Validate data model. """
        if self.id is not None:
            validate_json_rpc_id(self.id, JsonRpcInternalError)

        if (self.result is None) == (self.error is None):
            raise JsonRpcInternalError(
                "Response must contain one of `result` or `error`."
            )

        if self.result and not isinstance(
            self.result, (int, float, bool, str, list, dict)
        ):
            raise JsonRpcInternalError("`result` must be a valid JSON type.")

        if self.error and not isinstance(self.error, JsonRpcError):
            raise JsonRpcInternalError("`error` must be a JsonRpcError.")

    @property
    def success(self):
        """ True if the response contains an error. """
        return self.error is None

    def to_json_dict(self) -> JsonDict:
        """ Convert to a JSON dictionary. """
        dict_ = typing.cast(JsonDict, {"id": self.id, "jsonrpc": self.jsonrpc})
        if self.success:
            dict_["result"] = self.result
        else:
            dict_["error"] = typing.cast(JsonRpcError, self.error).to_json_dict()
        return dict_

    @classmethod
    def from_json_dict(cls, json_dict: JsonDict) -> JsonRpcResponse:
        """ Return a new response from a JSON dictionary. """
        error: typing.Optional[JsonRpcError]
        if "error" in json_dict:
            error = JsonRpcError.from_json_dict(
                typing.cast(JsonDict, json_dict["error"])
            )
        else:
            error = None
        result = json_dict["result"] if "result" in json_dict else None
        return cls(
            id=typing.cast(typing.Union[int, str, None], json_dict["id"]),
            error=error,
            result=result,
            jsonrpc=typing.cast(str, json_dict["jsonrpc"]),
        )


# This type is not defined in the types module because it relies on the class above.
JsonRpcResponseCallback = typing.Callable[[JsonRpcResponse], None]


class JsonRpcPeer:
    """
    Represents a JSON RPC client or server.

    The specification distinguishes between client and server but then adds that the
    implementation may "fill both of those roles, even at the same time, to...the same
    client." The wording is imprecise, but it suggests that the most flexible way to
    implement JSON RPC is a combined client/server object.

    For example, the specification states that notifications can only be sent from
    client to server, but it would be helpful to send from server to client as well. In
    order to fit the spec we have to interpret the client and server as temporarily
    switching roles.
    """

    def __init__(
        self, request_handler: typing.Optional[JsonRpcRequestHandler] = None,
    ):
        """ Constructor """
        self._id_gen = itertools.count()

    def request(
        self, method: str, params: typing.Optional[JsonRpcParams] = None,
    ) -> typing.Tuple[JsonRpcId, bytes]:
        """
        Create a new request.

        :param method: The method to invoke on the JSON-RPC server.
        :param params: Parameters to pass to the remote method.
        """
        request_id = next(self._id_gen)
        req = JsonRpcRequest(id=request_id, method=method, params=params)
        bytes_to_send = json.dumps(req.to_json_dict()).encode("utf8")
        return request_id, bytes_to_send

    def notify(
        self, method: str, params: typing.Optional[JsonRpcParams] = None
    ) -> bytes:
        """ Create a notification and return a network representation. """
        req = JsonRpcRequest(id=MissingId(), method=method, params=params)
        return json.dumps(req.to_json_dict()).encode("utf8")

    def respond_with_result(
        self, request: JsonRpcRequest, result: JsonPrimitive
    ) -> bytes:
        """
        Create a success response to a given request and return a network
        representation.
        """
        resp = JsonRpcResponse(id=typing.cast(JsonRpcId, request.id), result=result)
        return json.dumps(resp.to_json_dict()).encode("utf8")

    def respond_with_error(
        self, request: typing.Optional[JsonRpcRequest], error: JsonRpcError
    ) -> bytes:
        """
        Create an error response to a given request and return a network representation.

        :param request: If a request ID could be parsed, pass the request object.
            Otherwise pass None.
        :param error: The error information to respond with.
        """
        # If the request could not be parsed
        request_id: typing.Optional[JsonRpcId]
        if request is None:
            request_id = None
        else:
            request_id = typing.cast(JsonRpcId, request.id)
        resp = JsonRpcResponse(id=request_id, error=error)
        return json.dumps(resp.to_json_dict()).encode("utf8")

    def parse(
        self, recv_bytes: bytes
    ) -> typing.Iterable[typing.Union[JsonRpcRequest, JsonRpcResponse]]:
        """
        Parse a network representation.

        :returns: an iterable of parsed objects
        :raises JsonRpcParseError: if the data cannot be parsed
        """

        try:
            recv_str = recv_bytes.decode("utf8")
        except Exception:
            raise JsonRpcParseError("Invalid ASCII encoding")

        try:
            recv_dict = json.loads(recv_str)
        except:
            raise JsonRpcParseError("Invalid JSON format")

        messages: typing.Iterable[typing.Union[JsonRpcRequest, JsonRpcResponse]]

        if "method" in recv_dict:
            messages = (JsonRpcRequest.from_json_dict(recv_dict),)
        elif "result" in recv_dict or "error" in recv_dict:
            messages = (JsonRpcResponse.from_json_dict(recv_dict),)
        else:
            msg = "Could parse a request or a response: "
            example = recv_str[:100] + ("..." if len(recv_str) > 100 else "")
            raise JsonRpcParseError(msg + example)

        return messages



import typing

JsonRpcId = typing.Union[int, float, str, None]
# Note that the union should contain JsonDict instead of dict, and JsonList instead of
# list, but recursive types are not currently supported in MyPy.
JsonPrimitive = typing.Union[int, float, str, dict, list, None]
JsonDict = typing.Dict[str, JsonPrimitive]
JsonList = typing.List[JsonPrimitive]
JsonRpcParams = typing.Union[JsonDict, JsonList]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sansio_jsonrpc']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'sansio-jsonrpc',
    'version': '0.2.0',
    'description': 'JSON-RPC v2.0 Sans I/O',
    'long_description': '# JSON-RPC v2.0 Sans I/O\n\n[![PyPI](https://img.shields.io/pypi/v/sansio-jsonrpc.svg?style=flat-square)](https://pypi.org/project/sansio-jsonrpc/)\n![Python Versions](https://img.shields.io/pypi/pyversions/sansio-jsonrpc.svg?style=flat-square)\n![MIT License](https://img.shields.io/github/license/HyperionGray/sansio-jsonrpc.svg?style=flat-square)\n[![Build Status](https://img.shields.io/travis/com/HyperionGray/sansio-jsonrpc.svg?style=flat-square&branch=master)](https://travis-ci.com/HyperionGray/sansio-jsonrpc)\n[![codecov](https://codecov.io/gh/HyperionGray/sansio-jsonrpc/branch/master/graph/badge.svg)](https://codecov.io/gh/HyperionGray/sansio-jsonrpc)\n\nThis project provides [a Sans I/O](https://sans-io.readthedocs.io/) implementation of\n[JSON-RPC v 2.0](https://www.jsonrpc.org/specification). This means that the library\nhandles all of the encoding, decoding, framing, and logic required by the protocol\nspecification, but it does not implement any I/O (input or output). In order to use this\nlibrary, you need to either to use a downstream project that wraps I/O around this\nproject or else write your own I/O wrapper.\n\n## Client Example\n\nThis example illustrates what the library does and how it needs to be integrated into\nan I/O framework of your choosing.\n\n```python\nfrom sansio_jsonrpc import JsonRpcPeer, JsonRpcParseError, JsonRpcException\n\nclient = JsonRpcPeer()\nrequest_id, bytes_to_send = client.request(\n    method=\'open_vault_door\',\n    args={\'employee\': \'Mark\', \'pin\': 1234},\n)\n```\n\nFirst, we instantiate the client. We call the `request` method to create a new JSON-RPC\nrequest and convert it into a `bytes` representation that is suitable for sending to the\nserver. The method also returns a `request_id`, which is a hashable value that can be\nused to correlate any future response back to the request that initiated it.\n\n```python\nconnection.send(bytes_to_send)\nreceived_bytes = connection.recv()\n```\n\nRemember, this is a Sans I/O library, so it does not do any networking itself. It is the\ncaller\'s responsibility to transfer data to and from the remote machine. This block\nshows a hypothetical I/O framework being used to send the pending request and receive\nthe response as `bytes`, but this can be implemented with any I/O framework of your\nchoosing.\n\n```python\ntry:\n    messages = client.parse(received_bytes)\nexcept JsonRpcParseError as pe:\n    print(\'Exception while parsing response\', pe)\n```\n\nIn this block, we feed the data received from the remote machine into the client\nobject\'s `parse()` method. This method parses the incoming data and returns an iterable\ncontaining messages sent by the server. (In the current implementation, the iterable\nalways contains exactly 1 message, but the API is designed this way to allow for future\nenhancements such as a streaming JSON parser that can return 0-n messages after parsing\neach chunk of data.)\n\n```python\nfor response in messages:\n    assert isinstance(response, JsonRpcResponse)\n    print(\'received a response:\', response.id)\n    if response.is_success:\n        print(\'success:\', response.result)\n    else:\n        print(\'error:\', response.error)\n        exc = JsonRpcException.exc_from_error(response.error)\n```\n\nFinally, we iterate over the messages. In the case of a client, these messages will\nalways be `JsonRpcReponse` objects, which contain either result or error information.\nThe `response.id` field should match the `request_id` obtained earlier.\n\nIf you want an error response to raise an exception, you can call\n`JsonRpcException.exc_from_error(...)` and it will construct an exception for you. The\nexception system is described in a separate section below. Of course, be careful where\nyou throw the exception, since throwing an exception while reading multiple responses\nwill result in the rest of the responses being ignored. In a concurrent system, you\nprobably want to raise the exception to the code path that called `request()`.\n\nIn a concurrent system, you will also want to multiplex requests and responses on the\nsame connection. This library does not implement multiplexing itself because\nmultiplexing with appropriate flow control (see "Back Pressure" below) is going to\ndepend on the transport and I/O framework you are using.\n\n## Server Example\n\nThis example shows how to receive a JSON-RPC request and dispatch a response.\n\n```python\nfrom sansio_jsonrpc import (\n    JsonRpcPeer,\n    JsonRpcException,\n    JsonRpcParseError,\n    JsonRpcMethodNotFoundError,\n)\n\nserver = JsonRpcPeer()\n```\n\nThis example starts out the same way as the client: instantiating a `JsonRpcPeer`\nobject. Note that the client and server implementation are actually in the same class!\n\nThis might be surprising at first but it is the most flexible way to implement JSON-RPC.\nFor example, the specification says that a notification is a special type of request,\nand a request can only be sent from client to server. But there are scenarios where it\nwould be useful to have the server send notifications to the client, e.g. a\npublish/subscribe system. The specification says that:\n\n> One implementation of this specification could easily fill both of those roles, even\n> at the same time, to other different clients or the same client.\n\nThis is why the library implements both roles in a single class. The choice to name the\nobject `server` is just a convention to help us remember what\'s going on.\n\n```python\nreceived_bytes = connection.recv()\n```\n\nThis block shows the hypothetical I/O framework again. As a server, we just want to wait\nfor a client to send us something.\n\n```python\ntry:\n    messages = server.parse(received_bytes)\nexcept JsonRpcParseError as pe:\n    bytes_to_send = server.respond_with_error(request=None, error=pe.get_error())\n    connection.send(bytes_to_send)\n    return\n```\n\nNext, we want to parse the incoming data. If a parse error is raised here, then we\nshould send back a response and we should not iterate over the messages, because\n`messages` was not actually returned!. Because we were unable to parse a request, we set\n`request=None` in the response.\n\n```\nfor request in messages:\n    assert isinstance(request, JsonRpcRequest)\n\n    try:\n        # Handle request here and send a response.\n        result = handle_request(request)\n        bytes_to_send = server.respond_with_result(request=request, result=result)\n    except JsonRpcException as jre:\n        bytes_to_send = server.respond_with_error(request=request, error=pe.get_error())\n\n    connection.send(bytes_to_send)\n```\n\nNext, we iterate over the received messages. For a server, each message should be a\n`JsonRpcRequest`. The implementation of `handle_request(...)` is up to you! You should\nhandle each request by returning a result. The `respond_with_result(...)` method\nproduces an appropriate JSON-RPC response, which you should then send to the client.\n\nIf an error occurs in your handler, you should raise one of the built-in exceptions or a\ncustom subclass of `JsonRpcApplicationError`. For example, if you don\'t have a handler\nfor a given method, you should raise `JsonRpcMethodNotFoundError`. Exceptions are\ndescribed more fully below.\n\n## Exceptions\n\nThe exception system in this library is designed to make error-handling as Pythonic as\npossible, abstracting over the details of the JSON-RPC protocol as much as possible. All\nexceptions in this module inherit from `JsonRpcException`, so this is a good choice for\na catch-all exception handler. All exceptions have an error code and a message.\nExceptions can also optionally have a `data` field which can be set to any valid JSON\ndata type.\n\nThere are also specialized errors in the library that correspond to specific JSON-RPC\nerror codes in the specification. For example, if you receive data that cannot be\ndecoded into a valid JSON-RPC request, the specification calls for a -32700 error, and\nthis library raises a `JsonRpcParseError` to make it easy to catch this specific\ncondition. If the server sends\n\nThe specification also allows you to define your own error codes. There are\n\nExceptions are used in two ways in this library. First of all, the library itself can\nraise exceptions. For example in `recv()`, it can raise `JsonRpcParseError` as described\nabove. The second usage is that the remote peer might send you a response that contains\nan error object. In this case, you may want to convert that error into an exception so\nyou can raise it.\n\nThe static method `JsonRpcException.exc_from_error(...)` converts an error into an\nexception. This method automatically selects the appropriate subclass of\n`JsonRpcException`. For example, if the remote peer sent you error code -32700, this\nmethod converts that into a `JsonRpcParseError`. Some error codes are reserved for\nimplementations to define their own error codes, e.g. -32099 is reserved in the\nspecification but not assigned any meaning. If the remote peer sends this error code,\nthen this method will raise `JsonRpcReservedError`.\n\nThe specification also has unreserved error codes that are available for defining\napplication-specific errors. You can use these in your own code in two different ways.\nThe first approach is to catch/raise `JsonRpcApplicationError` and set/check the error\ncode to your application-defined value. The second approach is to create your own\nsubclass of `JsonRpcApplicationError`. Here\'s an example:\n\n```python\nfrom sansio_jsonrpc import JsonRpcApplicationError\n\n\nclass MyApplicationError1(JsonRpcApplicationError):\n    ERROR_CODE = 1\n    ERROR_MESSAGE = "My application error type 1"\n\n\nclass MyApplicationError2(JsonRpcApplicationError):\n    ERROR_CODE = 2\n    ERROR_MESSAGE = "My application error type 2"\n```\n\nThe `JsonRpcException.exc_from_error(...)` method will automatically select the\nappropriate subclass. For example if the server sends error code 1, then the method will\nreturn a `MyApplicationError1` instance, even though that class is defined in _your\ncode_! The library uses some metaclass black magic to make this work.\n\n## Back Pressure\n\nAs a SANS I/O library, this package does not implement any sort of flow control. If the\npeer is sending you data faster than you can handle it, and you keep reading it, then\nyour process\'s memory usage will continually grow until it runs out of memory or the\nkernel terminates it. Back pressure means signalling to the peer that it should stop\nsending data for a bit until your process can catch up. The specifics of back pressure\nreally depend on what transport protocol and I/O framework you are using. For example,\nTCP has flow control capabilities, and most implementations will automatically apply\nback pressure if you simply stop reading from the socket. Therefore, if you use TCP with\nthis library, you should be careful to read from the socket only when you are ready to\nprocess another message. If you eagerly read from the socket into an unbounded user\nspace buffer (such as a queue), then your code will not benefit from TCP\'s flow control,\nbecause the kernel will see an empty buffer and it will keep filling it up.\n\n## Developing\n\nThe project uses MyPy for type checking and Black for code formatting. Poetry is used to\nmanage dependencies and to build releases.\n',
    'author': 'Mark E. Haase',
    'author_email': 'mehaase@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/hyperiongray/sansio-jsonrpc',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
